#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2017 by Christian Tremblay, P.Eng <christian.tremblay@servisysDeviceObject.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Device.py - describe a BACnet Device

"""
# --- standard Python modules ---
from collections import namedtuple
from datetime import datetime
import weakref

import os.path

# --- 3rd party modules ---
import sqlite3

try:
    import pandas as pd

    _PANDAS = True
except ImportError:
    _PANDAS = False
import logging

try:
    from xlwings import Workbook, Sheet, Range, Chart

    _XLWINGS = True
except ImportError:
    _XLWINGS = False


# --- this application's modules ---
from bacpypes.basetypes import ServicesSupported

from .Points import NumericPoint, BooleanPoint, EnumPoint, OfflinePoint
from ..io.IOExceptions import (
    NoResponseFromController,
    SegmentationNotSupported,
    BadDeviceDefinition,
    RemovedPointException,
    WritePropertyException,
    WrongParameter,
    DeviceNotConnected,
)

# from ...bokeh.BokehRenderer import BokehPlot
from ...sql.sql import SQLMixin
from ...tasks.DoOnce import DoOnce
from .mixins.read_mixin import ReadPropertyMultiple, ReadProperty

from ..utils.notes import note_and_log


# ------------------------------------------------------------------------------


class DeviceProperties(object):
    """
    This serves as a container for device properties
    """

    def __init__(self):
        self.name = "Unknown"
        self.address = None
        self.device_id = None
        self.network = None
        self.pollDelay = None
        self.objects_list = None
        self.pss = ServicesSupported()
        self.multistates = None
        self.db_name = None
        self.segmentation_supported = True
        self.history_size = None
        self.bacnet_properties = {}

    def __repr__(self):
        return "{}".format(self.asdict)

    @property
    def asdict(self):
        return self.__dict__


@note_and_log
class Device(SQLMixin):
    """
    Represent a BACnet device.  Once defined, it allows use of read, write, sim, release
    functions to communicate with the device on the network.

    :param address: address of the device (ex. '2:5')
    :param device_id: bacnet device ID (boid)
    :param network: defined by BAC0.connect()
    :param poll: (int) if > 0, will poll every points each x seconds.
    :from_backup: sqlite backup file
    :segmentation_supported: (boolean) When segmentation is not supported, BAC0
                             will not use read property multiple to poll the
                             device.
    :object_list: (list) Use can provide a custom object_list to use for the
                  the creation of the device. the object list must be built
                  using the same pattern returned by bacpypes when polling the
                  objectList property
                  example ::
                      my_obj_list = [('file', 1),
                     ('analogInput', 2),
                     ('analogInput', 3),
                     ('analogInput', 5),
                     ('analogInput', 4),
                     ('analogInput', 0),
                     ('analogInput', 1)]
    :auto_save: (False or int) If False or 0, auto_save is disabled. To
                Activate, pass an integer representing the number of polls
                before auto_save is called. Will write the histories to
                SQLite db locally.
    :clear_history_on_save: (boolean) Will clear device history

    :type address: (str)
    :type device_id: int
    :type network: BAC0.scripts.ReadWriteScript.ReadWriteScript
    """

    def __init__(
        self,
        address=None,
        device_id=None,
        network=None,
        *,
        poll=10,
        from_backup=None,
        segmentation_supported=True,
        object_list=None,
        auto_save=False,
        save_resampling="1s",
        clear_history_on_save=False,
        history_size=None,
        reconnect_on_failure=True,
    ):

        self.properties = DeviceProperties()

        self.properties.address = address
        self.properties.device_id = device_id
        self.properties.network = network
        self.properties.pollDelay = poll
        self.properties.fast_polling = True if poll < 10 else False
        self.properties.name = ""
        self.properties.vendor_id = 0
        self.properties.objects_list = []
        self.properties.pss = ServicesSupported()
        self.properties.multistates = {}
        self.properties.auto_save = auto_save
        self.properties.save_resampling = save_resampling
        self.properties.clear_history_on_save = clear_history_on_save
        self.properties.default_history_size = history_size
        self._reconnect_on_failure = reconnect_on_failure

        self.segmentation_supported = segmentation_supported
        self.custom_object_list = object_list

        # self.db = None
        # Todo : find a way to normalize the name of the db
        self.properties.db_name = ""

        self.points = []
        self._list_of_trendlogs = {}

        self._polling_task = namedtuple("_polling_task", ["task", "running"])
        self._polling_task.task = None
        self._polling_task.running = False

        self._find_overrides_running = False
        self._release_overrides_running = False

        self.note("Controller initialized")

        if from_backup:
            filename = from_backup
            db_name = filename.split(".")[0]
            self.properties.network = None
            if os.path.isfile(filename):
                self.properties.db_name = db_name
                self.new_state(DeviceDisconnected)
            else:
                raise FileNotFoundError("Can't find {} on drive".format(filename))
        else:
            if (
                self.properties.network
                and self.properties.address
                and self.properties.device_id is not None
            ):
                self.new_state(DeviceDisconnected)
            else:
                raise BadDeviceDefinition(
                    "Please provide address, device id and network or specify from_backup argument"
                )

    def new_state(self, newstate):
        """
        Base of the state machine mechanism.
        Used to make transitions between device states.
        Take care to call the state init function.
        """
        self._log.info(
            "Changing device state to {}".format(str(newstate).split(".")[-1])
        )
        self.__class__ = newstate
        self._init_state()

    def _init_state(self):
        """
        Execute additional code upon state modification
        """
        raise NotImplementedError()

    def connect(self):
        """
        Connect the device to the network
        """
        raise NotImplementedError()

    def disconnect(self):
        raise NotImplementedError()

    def initialize_device_from_db(self):
        raise NotImplementedError()

    def df(self, list_of_points, force_read=True):
        """
        Build a pandas DataFrame from a list of points.  DataFrames are used to present and analyze data.

        :param list_of_points: a list of point names as str
        :returns: pd.DataFrame
        """
        raise NotImplementedError()

    @property
    def simulated_points(self):
        """
        iterate over simulated points

        :returns: points if simulated (out_of_service == True)
        :rtype: BAC0.core.devices.Points.Point
        """
        for each in self.points:
            if each.properties.simulated:
                yield each

    def _buildPointList(self):
        """
        Read all points from a device into a (Pandas) dataframe (Pandas).  Items are
        accessible by point name.
        """
        raise NotImplementedError()

    def __getitem__(self, point_name):
        """
        Get a point from its name.
        If a list is passed - a dataframe is returned.

        :param point_name: (str) name of the point or list of point_names
        :type point_name: str
        :returns: (Point) the point (can be Numeric, Boolean or Enum) or pd.DataFrame
        """
        raise NotImplementedError()

    def __iter__(self):
        """
        When iterating a device, iterate points of it.
        """
        raise NotImplementedError()

    def __contains__(self, value):
        "When using in..."
        raise NotImplementedError()

    @property
    def points_name(self):
        """
        When iterating a device, iterate points of it.
        """
        raise NotImplementedError()

    def to_excel(self):
        """
        Using xlwings, make a dataframe of all histories and save it
        """
        raise NotImplementedError()

    def __setitem__(self, point_name, value):
        """
        Write, sim or ovr value

        :param point_name: Name of the point to set
        :param value: value to write to the point
        :type point_name: str
        :type value: float
        """
        raise NotImplementedError()

    def __len__(self):
        """
        Will return number of points available
        """
        raise NotImplementedError()

    def _parseArgs(self, arg):
        """
        Given a string, interpret the last word as the value, everything else is
        considered to be the point name.
        """
        args = arg.split()
        pointName = " ".join(args[:-1])
        value = args[-1]
        return (pointName, value)

    def clear_histories(self):
        for point in self.points:
            point.clear_history()

    def update_history_size(self, size=None):
        for point in self.points:
            point.properties.history_size = size

    @property
    def analog_units(self):
        raise NotImplementedError()

    @property
    def temperatures(self):
        raise NotImplementedError()

    @property
    def percent(self):
        raise NotImplementedError()

    @property
    def multi_states(self):
        raise NotImplementedError()

    @property
    def binary_states(self):
        raise NotImplementedError()

    def _findPoint(self, name, force_read=True):
        """
        Helper that retrieve point based on its name.

        :param name: (str) name of the point
        :param force_read: (bool) read value of the point each time the function is called.
        :returns: Point object
        :rtype: BAC0.core.devices.Point.Point (NumericPoint, EnumPoint or BooleanPoint)
        """
        raise NotImplementedError()

    def find_point(self, objectType, objectAddress):
        """
        Find point based on type and address
        """
        for point in self.points:
            if (
                point.properties.type == objectType
                and float(point.properties.address) == objectAddress
            ):
                return point
        raise ValueError(
            "{} {} doesn't exist in controller".format(objectType, objectAddress)
        )

    def find_overrides(self, force=False):
        if self._find_overrides_running and not force:
            self._log.warning(
                "Already running ({:.1%})... please wait.".format(
                    self._find_overrides_progress
                )
            )
            return
        lst = []
        self._find_overrides_progress = 0
        self._find_overrides_running = True
        total = len(self.points)

        def _find_overrides():
            self._log.warning(
                "Overrides are being checked, wait for completion message."
            )
            for idx, point in enumerate(self.points):
                if point.is_overridden:
                    lst.append(point)
                self._find_overrides_progress = idx / total
            self._log.warning(
                "Override check ready, results available in device.properties.points_overridden"
            )
            self.properties.points_overridden = lst
            self._find_overrides_running = False
            self._find_overrides_progress = 1

        self.do(_find_overrides)

    def find_overrides_progress(self):
        return self._find_overrides_progress

    def release_all_overrides(self, force=False):
        if self._release_overrides_running and not force:
            self._log.warning(
                "Already running ({:.1%})... please wait.".format(
                    self._release_overrides_progress
                )
            )
            return
        self._release_overrides_running = True
        self._release_overrides_progress = 0

        def _release_all_overrides():
            self.find_overrides()
            while self._find_overrides_running:
                self._release_overrides_progress = self._find_overrides_progress * 0.5

            if self.properties.points_overridden:
                total = len(self.properties.points_overridden)
                self._log.info("=================================")
                self._log.info("Overrides found... releasing them")
                self._log.info("=================================")
                for idx, point in enumerate(self.properties.points_overridden):
                    self._log.info("Releasing {}".format(point))
                    point.release_ovr()
                    self._release_overrides_progress = (idx / total) / 2 + 0.5
            else:
                self._log.info("No override found")

            self._release_overrides_running = False
            self._release_overrides_progress = 1

        self.do(_release_all_overrides)

    def do(self, func):
        DoOnce(func).start()

    def __repr__(self):
        return "{} / Undefined".format(self.properties.name)


# @fix_docs
class DeviceConnected(Device):
    """
    Find a device on the BACnet network.  Set its state to 'connected'.
    Once connected, all subsequent commands use this BACnet connection.
    """

    def _init_state(self):
        self._buildPointList()
        self.properties.network.register_device(self)

    def disconnect(self, save_on_disconnect=True, unregister=True):
        self._log.info("Wait while stopping polling")
        self.poll(command="stop")
        if save_on_disconnect:
            self.save()
        if unregister:
            self.properties.network.unregister_device(self)
        self.new_state(DeviceFromDB)

    def connect(self, *, db=None):
        """
        A connected device can be switched to 'database mode' where the device will
        not use the BACnet network but instead obtain its contents from a previously
        stored database.
        """
        if db:
            self.poll(command="stop")
            self.properties.db_name = db.split(".")[0]
            self.new_state(DeviceFromDB)
        else:
            self._log.warning(
                "Already connected, provide db arg if you want to connect to db"
            )

    def df(self, list_of_points, force_read=True):
        """
        When connected, calling DF should force a reading on the network.
        """

        his = []
        for point in list_of_points:
            try:
                his.append(self._findPoint(point, force_read=force_read).history)
            except ValueError as ve:
                self._log.error("{}".format(ve))
                continue
        if not _PANDAS:
            return dict(zip(list_of_points, his))
        return pd.DataFrame(dict(zip(list_of_points, his)))

    def _buildPointList(self):
        """
        Upon connection to build the device point list and properties.
        """
        try:
            self.properties.pss.value = self.properties.network.read(
                "{} device {} protocolServicesSupported".format(
                    self.properties.address, self.properties.device_id
                )
            )

        except NoResponseFromController as error:
            self._log.error("Controller not found, aborting. ({})".format(error))
            return ("Not Found", "", [], [])

        except SegmentationNotSupported as error:
            self._log.warning("Segmentation not supported")
            self.segmentation_supported = False
            self.new_state(DeviceDisconnected)

        self.properties.name = self.properties.network.read(
            "{} device {} objectName".format(
                self.properties.address, self.properties.device_id
            )
        )
        self.properties.vendor_id = self.properties.network.read(
            "{} device {} vendorIdentifier".format(
                self.properties.address, self.properties.device_id
            )
        )
        self._log.info(
            "Device {}:[{}] found... building points list".format(
                self.properties.device_id, self.properties.name
            )
        )
        try:
            self.properties.objects_list, self.points, self._list_of_trendlogs = self._discoverPoints(
                self.custom_object_list
            )
            if self.properties.pollDelay > 0:
                self.poll(delay=self.properties.pollDelay)
            self.update_history_size(size=self.properties.default_history_size)
            # self.clear_histories()
        except NoResponseFromController as error:
            self._log.error("Cannot retrieve object list, disconnecting...")
            self.segmentation_supported = False
            self.new_state(DeviceDisconnected)
        except IndexError as error:
            if self._reconnect_on_failure:
                self._log.error("Device creation failed... re-connecting")
                self.new_state(DeviceDisconnected)
            else:
                self._log.error("Device creation failed... disconnecting")

    def __getitem__(self, point_name):
        """
        Allows the syntax: device['point_name'] or device[list_of_points]

        If calling a list, last value will be used (won't read on the network)
        for performance reasons.
        If calling a simple point, point will be read via BACnet.
        """
        try:
            if isinstance(point_name, list):
                return self.df(point_name, force_read=False)
            elif isinstance(point_name, tuple):
                _type, _address = point_name
                for point in self.points:
                    if point.properties.type == _type and str(
                        point.properties.address
                    ) == str(_address):
                        return point
            else:
                try:
                    return self._findPoint(point_name, force_read=False)
                except ValueError:
                    try:
                        return self._findTrend(point_name)
                    except ValueError:
                        try:
                            if "@prop_" in point_name:
                                point_name = point_name.split("prop_")[1]
                                return self.read_property(
                                    ("device", self.properties.device_id, point_name)
                                )
                            else:
                                raise ValueError()
                        except ValueError as ve:
                            raise ValueError()
        except ValueError as ve:
            self._log.error("{}".format(ve))

    def __iter__(self):
        yield from self.points

    def __contains__(self, value):
        """
        Allows the syntax:
            if "point_name" in device:
        """
        return value in self.points_name

    @property
    def points_name(self):
        for each in self.points:
            yield each.properties.name

    def __setitem__(self, point_name, value):
        """
        Allows the syntax:
            device['point_name'] = value
        """
        try:
            self._findPoint(point_name)._set(value)
        except WritePropertyException as ve:
            self._log.error("{}".format(ve))

    def __len__(self):
        """
        Length of a device = number of points
        """
        return len(self.points)

    def _parseArgs(self, arg):
        args = arg.split()
        pointName = " ".join(args[:-1])
        value = args[-1]
        return (pointName, value)

    @property
    def analog_units(self):
        """
        Shortcut to retrieve all analog points units [Used by Bokeh trending feature]
        """
        au = []
        us = []
        for each in self.points:
            if isinstance(each, NumericPoint):
                au.append(each.properties.name)
                us.append(each.properties.units_state)
        return dict(zip(au, us))

    @property
    def temperatures(self):
        for each in self.analog_units.items():
            if "deg" in each[1]:
                yield each

    @property
    def percent(self):
        for each in self.analog_units.items():
            if "percent" in each[1]:
                yield each

    @property
    def multi_states(self):
        ms = []
        us = []
        for each in self.points:
            if isinstance(each, EnumPoint):
                ms.append(each.properties.name)
                us.append(each.properties.units_state)
        return dict(zip(ms, us))

    @property
    def binary_states(self):
        bs = []
        us = []

        for each in self.points:
            if isinstance(each, BooleanPoint):
                bs.append(each.properties.name)
                us.append(each.properties.units_state)
        return dict(zip(bs, us))

    def _findPoint(self, name, force_read=False):
        """
        Used by getter and setter functions
        """
        for point in self.points:
            if point.properties.name == name:
                if force_read:
                    point.value
                return point
        raise ValueError("{} doesn't exist in controller".format(name))

    def _trendlogs(self):
        for k, v in self._list_of_trendlogs.items():
            name, trendlog = v
            yield trendlog

    @property
    def trendlogs_names(self):
        for each in self._trendlogs():
            yield each.properties.object_name

    @property
    def trendlogs(self):
        return list(self._trendlogs())

    def _findTrend(self, name):
        for trend in self._trendlogs():
            if trend.properties.object_name == name:
                return trend
        raise ValueError("{} doesn't exist in controller".format(name))

    def read_property(self, prop):
        # if instance == -1:
        #    pass
        if isinstance(prop, tuple):
            _obj, _instance, _prop = prop
        elif isinstance(prop, str):
            _obj = "device"
            _instance = self.properties.device_id
            _prop = prop
        else:
            raise ValueError(
                "Please provide property using tuple with object, instance and property"
            )
        try:
            request = "{} {} {} {}".format(
                self.properties.address, _obj, _instance, _prop
            )
            val = self.properties.network.read(request, vendor_id=0)
        except KeyError as error:
            raise Exception("Unknown property : {}".format(error))
        return val

    def write_property(self, prop, value, priority=None):
        if priority is not None:
            priority = "- {}".format(priority)
        if isinstance(prop, tuple):
            _obj, _instance, _prop = prop
        else:
            raise ValueError(
                "Please provide property using tuple with object, instance and property"
            )
        try:
            request = "{} {} {} {} {} {}".format(
                self.properties.address, _obj, _instance, _prop, value, priority
            )
            val = self.properties.network.write(
                request, vendor_id=self.properties.vendor_id
            )
        except KeyError as error:
            raise Exception("Unknown property : {}".format(error))
        return val

    def update_bacnet_properties(self):
        """
        Retrieve bacnet properties for this device
        To retrieve something general, forcing vendor id 0
        """
        try:
            res = self.properties.network.readMultiple(
                "{} device {} all".format(
                    self.properties.address, str(self.properties.device_id)
                ),
                vendor_id=0,
                prop_id_required=True,
            )
            for each in res:
                if not each:
                    continue
                v, prop = each
                self.properties.bacnet_properties[prop] = v

        except Exception as e:
            raise Exception("Problem reading : {} | {}".format(self.properties.name, e))

    def _bacnet_properties(self, update=False):
        if not self.properties.bacnet_properties or update:
            self.update_bacnet_properties()
        return self.properties.bacnet_properties

    @property
    def bacnet_properties(self):
        return self._bacnet_properties(update=True)

    def __repr__(self):
        return "{} / Connected".format(self.properties.name)


# ------------------------------------------------------------------------------


class RPDeviceConnected(DeviceConnected, ReadProperty):
    """
    [Device state] If device is connected but doesn't support ReadPropertyMultiple

    BAC0 will not poll such points automatically (since it would cause excessive network traffic).
    Instead manual polling must be used as needed via the poll() function.
    """

    def __str__(self):
        return "connected [for ReadProperty]"


class RPMDeviceConnected(DeviceConnected, ReadPropertyMultiple):
    """
    [Device state] If device is connected and supports ReadPropertyMultiple
    """

    def __str__(self):
        return "connected [for ReadPropertyMultiple]"


# @fix_docs
class DeviceDisconnected(Device):
    """
    [Device state] Initial state of a device. Disconnected from BACnet.
    """

    def _init_state(self):
        self.connect()

    def connect(self, *, db=None):
        """
        Attempt to connect to device.  If unable, attempt to connect to a controller database
        (so the user can use previously saved data).
        """
        if not self.properties.network:
            self._log.debug("No network...calling DeviceFromDB")
            self.new_state(DeviceFromDB)
        else:
            try:
                name = self.properties.network.read(
                    "{} device {} objectName".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                segmentation = self.properties.network.read(
                    "{} device {} segmentationSupported".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                if not self.segmentation_supported or segmentation not in (
                    "segmentedTransmit",
                    "segmentedBoth",
                ):
                    segmentation_supported = False
                    self._log.debug("Segmentation not supported")
                else:
                    segmentation_supported = True

                if name:
                    if segmentation_supported:
                        self.new_state(RPMDeviceConnected)
                    else:
                        self.new_state(RPDeviceConnected)

            except SegmentationNotSupported:
                self.segmentation_supported = False
                self._log.warning(
                    "Segmentation not supported.... expect slow responses."
                )
                self.new_state(RPDeviceConnected)

            except (NoResponseFromController, AttributeError) as error:
                if self.properties.db_name:
                    self.new_state(DeviceFromDB)
                else:
                    self._log.warning(
                        "Offline: provide database name to load stored data."
                    )
                    self._log.warning("Ex. controller.connect(db = 'backup')")

    def df(self, list_of_points, force_read=True):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def simulated_points(self):
        for each in self.points:
            if each.properties.simulated:
                yield each

    def _buildPointList(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    # This should be a "read" function and rpm defined in state rpm
    def read_multiple(
        self, points_list, *, points_per_request=25, discover_request=(None, 6)
    ):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def poll(self, command="start", *, delay=10):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __getitem__(self, point_name):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __iter__(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __contains__(self, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def points_name(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def to_excel(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __setitem__(self, point_name, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __len__(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def analog_units(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def temperatures(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def percent(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    @property
    def multi_states(self):
        raise DeviceNotConnected("Must connect to bacnet or database")

    @property
    def binary_states(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _discoverPoints(self, custom_object_list=None):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _findPoint(self, name, force_read=True):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __repr__(self):
        return "{} / Disconnected".format(self.properties.name)


# ------------------------------------------------------------------------------

# @fix_docs


class DeviceFromDB(DeviceConnected):
    """
    [Device state] Where requests for a point's present value returns the last
    valid value from the point's history.
    """

    def _init_state(self):
        try:
            self.initialize_device_from_db()
        except ValueError as e:
            self._log.error("Problem with DB initialization : {}".format(e))
            # self.new_state(DeviceDisconnected)
            raise

    def connect(self, *, network=None, from_backup=None):
        """
        In DBState, a device can be reconnected to BACnet using:
            device.connect(network=bacnet) (bacnet = BAC0.connect())
        """
        if network and from_backup:
            raise WrongParameter("Please provide network OR from_backup")

        elif network:
            self._log.debug("Network provided... trying to connect")
            self.properties.network = network
            try:
                name = self.properties.network.read(
                    "{} device {} objectName".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                segmentation = self.properties.network.read(
                    "{} device {} segmentationSupported".format(
                        self.properties.address, self.properties.device_id
                    )
                )

                if not self.segmentation_supported or segmentation not in (
                    "segmentedTransmit",
                    "segmentedBoth",
                ):
                    segmentation_supported = False
                    self._log.debug("Segmentation not supported")
                else:
                    segmentation_supported = True

                if name:
                    if segmentation_supported:
                        self._log.debug("Segmentation supported, connecting...")
                        self.new_state(RPMDeviceConnected)
                    else:
                        self._log.debug("Segmentation not supported, connecting...")
                        self.new_state(RPDeviceConnected)
                    # self.db.close()

            except NoResponseFromController:
                self._log.error("Unable to connect, keeping DB mode active")

        else:
            self._log.debug("Not connected, open DB")
            if from_backup:
                self.properties.db_name = from_backup.split(".")[0]
            self._init_state()

    def initialize_device_from_db(self):
        self._log.info("Initializing DB")
        # Save important properties for reuse
        if self.properties.db_name:
            dbname = self.properties.db_name
        else:
            self._log.info("Missing argument DB")
            raise ValueError("Please provide db name using device.load_db('name')")

        # network = self.properties.network
        pss = self.properties.pss

        self._props = self.read_dev_prop(self.properties.db_name)
        self.points = []
        for point in self.points_from_sql(self.properties.db_name):
            try:
                self.points.append(OfflinePoint(self, point))
            except RemovedPointException:
                continue

        self.properties = DeviceProperties()
        self.properties.db_name = dbname
        self.properties.address = self._props["address"]
        self.properties.device_id = self._props["device_id"]
        self.properties.network = None
        self.properties.pollDelay = self._props["pollDelay"]
        self.properties.name = self._props["name"]
        self.properties.objects_list = self._props["objects_list"]
        self.properties.pss = pss
        self.properties.serving_chart = {}
        self.properties.charts = []
        self.properties.multistates = self._props["multistates"]
        self._log.info("Device restored from db")
        self._log.info(
            'You can reconnect to network using : "device.connect(network=bacnet)"'
        )

    @property
    def simulated_points(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _buildPointList(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    # This should be a "read" function and rpm defined in state rpm
    def read_multiple(
        self, points_list, *, points_per_request=25, discover_request=(None, 6)
    ):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def poll(self, command="start", *, delay=10):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __contains__(self, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def to_excel(self):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __setitem__(self, point_name, value):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def _discoverPoints(self, custom_object_list=None):
        raise DeviceNotConnected("Must connect to BACnet or database")

    def __repr__(self):
        return "{} / Disconnected".format(self.properties.name)


# ------------------------------------------------------------------------------


class DeviceLoad(DeviceFromDB):
    def __init__(self, filename=None):
        if filename:
            Device.__init__(self, None, None, None, from_backup=filename)
        else:
            raise Exception("Please provide backup file as argument")



"""
Setup.py
"""
from setuptools import setup
from BAC0 import infos

requirements = ["bacpypes", "colorama"]

setup(
    name="BAC0",
    version=infos.__version__,
    description="BACnet Scripting Framework for testing DDC Controls",
    author=infos.__author__,
    author_email=infos.__email__,
    url=infos.__url__,
    download_url=infos.__download_url__,
    keywords=["bacnet", "building", "automation", "test"],
    packages=[
        "BAC0",
        "BAC0.core",
        "BAC0.core.app",
        "BAC0.core.io",
        "BAC0.core.functions",
        "BAC0.core.devices",
        "BAC0.core.devices.local",
        "BAC0.core.devices.mixins",
        "BAC0.core.utils",
        "BAC0.core.proprietary_objects",
        "BAC0.scripts",
        "BAC0.tasks",
        "BAC0.web",
        "BAC0.sql",
    ],
    include_package_data=True,
    requires=requirements,
    install_requires=requirements,
    test_suite="tests",
    long_description=open("README.rst").read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Networking",
        "Topic :: Utilities",
    ],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
TaskManager.py - creation of threads used for repetitive tasks.  

A key building block for point simulation.
"""
# --- standard Python modules ---
from threading import Thread, Lock
import time
from collections import deque
from random import random


# --- 3rd party modules ---
# --- this application's modules ---
from ..core.utils.notes import note_and_log
from ..core.io.IOExceptions import DeviceNotConnected

# ------------------------------------------------------------------------------


@note_and_log
class Manager:
    tasks = []
    manager = None
    enable = False

    def __init__(self):
        if not Manager.enable:
            self.start_service()

        self._log.debug("Task Manager Initiated")

    @classmethod
    def process(cls):
        task = None
        while cls.enable:
            try:
                if cls.tasks == []:
                    raise IndexError
                _temp = cls.tasks.copy()
                _temp.sort()
                if _temp[-1].next_execution <= time.time():
                    task = _temp.pop()
                    task.execute()
                    cls.tasks.remove(task.id)
                    task.previous_execution = time.time()
                    if task.delay > 0:
                        task.next_execution = task.previous_execution + task.delay
                        cls.schedule_task(task)

                    cls._log.debug(
                        "Task {} | {} executed. {}".format(task.id, task.name, task)
                    )
            except IndexError:
                cls._log.debug("Task Manager waiting for tasks...")
                time.sleep(1)
            except DeviceNotConnected as error:
                cls._log.warning(
                    "Device disconnected. Removing task ({}).".format(error, task)
                )
                cls.tasks.remove(task.id)
            except Exception as error:
                cls._log.error(
                    "Super Mega Giga big error {}. Removing task.".format(error)
                )
                cls.tasks.remove(task.id)
            else:
                if not cls.manager.is_alive() and cls.enable:
                    cls._log.error(
                        "TaskManager Thread stopped... This is not normal..."
                    )
                    cls.stop_service()
                    cls.start_service()
            time.sleep(0.01)
        cls.stop_service()

    @classmethod
    def schedule_task(cls, task):
        cls.tasks.append(task)

    @classmethod
    def stopAllTasks(cls):
        cls._log.info("Stopping all tasks")
        cls.enable = False
        while cls.manager.is_alive():
            time.sleep(0.01)

        cls._log.info("Ok all tasks stopped")
        cls.clean_tasklist()
        return True

    @classmethod
    def start_service(cls):
        cls._log.info("Starting TaskManager")
        cls.enable = True
        cls.manager = Thread(target=cls.process, daemon=True)
        cls.manager.start()

    @classmethod
    def stop_service(cls):
        cls._log.info("Stopping TaskManager")
        cls.enable = False
        # time.sleep(1)
        # cls.manager.join()

    @classmethod
    def clean_tasklist(cls):
        cls._log.debug("Cleaning tasks list")
        cls.tasks = []

    def __repr__(self):
        return "TaskManager"

    @classmethod
    def number_of_tasks(cls):
        return len(cls.tasks)


@note_and_log
class Task(object):
    _tasks = []

    def __init__(self, fn=None, name=None, delay=0):
        if not Manager.enable:
            _manager = Manager()
        # delay = 0 -> one shot
        if isinstance(fn, tuple):
            self.fn, self.args = fn
        else:
            self.fn = fn
            self.args = None
        self.name = name
        if delay > 0:
            self.delay = delay if delay >= 5 else 5
        else:
            self.delay = 0
        self.previous_execution = None
        self.average_execution_delay = 0
        self.average_latency = 0
        self.next_execution = time.time() + delay + (random() * 10)
        self.execution_time = 0
        self.count = 0
        self.id = id(self)
        self._kwargs = None
        Task._tasks.append(self)

    def task(self):
        raise NotImplementedError("Must be implemented")

    def execute(self):
        _start_time = time.time()
        self.count += 1
        self.average_latency = (
            self.average_latency + (_start_time - self.next_execution)
        ) / 2
        if self.fn and self.args is not None:
            self.fn(self.args)
        elif self.fn:
            self.fn()
        else:
            if self._kwargs is not None:
                self.task(self._kwargs)
            else:
                self.task()
        if self.previous_execution:
            _total = self.average_execution_delay + (
                _start_time - self.previous_execution
            )
            self.average_execution_delay = _total / 2
        else:
            self.average_execution_delay = self.delay

        # self._log.info('Stat for task {}'.format(self))
        if self.average_latency > 5:
            self._log.warning("High latency for {}".format(self.name))
            self._log.warning("Stats : {}".format(self))

        self._log.debug("Executing : {}".format(self.name))
        self.execution_time = time.time() - _start_time

    def start(self):
        Manager.schedule_task(task=self)

    def stop(self):
        self.delay = 0

    @property
    def last_time(self):
        return time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(self.previous_execution)
        )

    @property
    def next_time(self):
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.next_execution))

    @property
    def latency(self):
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.average_latency))

    def is_alive(self):
        return

    def __repr__(self):
        return "{:<40} | Avg exec delay : {:.2f} sec | Avg latency : {:.2f} sec | last executed : {} | Next Time : {}".format(
            self.name,
            self.average_execution_delay,
            self.average_latency,
            self.last_time,
            self.next_time,
        )

    def __lt__(self, other):
        # list sort use __lt__... little cheat to reverse list already
        return self.next_execution > other.next_execution

    def __eq__(self, other):
        # list remove use __eq__... so compare with id
        if isinstance(other, Task):
            return self.id == other.id
        else:
            return self.id == other


def stopAllTasks():
    return Manager.stopAllTasks()


@note_and_log
class OneShotTask(Task):
    def __init__(self, fn=None, args=None, name="Oneshot"):
        super().__init__(name=name, delay=0)



# --- standard Python modules ---
import weakref

# --- 3rd party modules ---
from bacpypes.core import deferred

# --- this application's modules ---
from .TaskManager import Task
from ..core.utils.notes import note_and_log


@note_and_log
class Update_local_COV(Task):
    """
    Start a task to validate each points inside local device and send 
    cov notifications to subscribers.
    """

    def __init__(self, device, delay=1, name="", prefix="update_local_cov"):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec
        
        A delay cannot be < 10sec 
        For delays under 10s, use DeviceFastPoll class.

        :returns: Nothing
        """
        self._device = weakref.ref(device)
        Task.__init__(self, name="{}_{}".format(prefix, name), delay=delay)
        self._counter = 0

    @property
    def device(self):
        return self._device()

    def task(self):
        try:
            for k, cov in self.device.this_application.cov_detections.items():
                objName = k.objectName
                obj = self.device.this_application.get_object_name(objName)
                # get the detection algorithm object
                cov_detection = self.device.this_application.cov_detections.get(
                    obj, None
                )
                if (not cov_detection) or (len(cov_detection.cov_subscriptions) == 0):
                    self._log.debug(
                        "no subscriptions for that object : {}".format(objName)
                    )
                    continue

                # tell it to send out notifications
                self._log.info("Sending COV for {}".format(objName))
                deferred(cov_detection.send_cov_notifications)

        except Exception as error:
            self.device._log.error(
                "Something is wrong with update_local_cov : {}".format(error)
            )



#!/usr/bin/env python

"""
Johnson Controls Proprietary Objects for FX/FEC Line
"""
from bacpypes.primitivedata import (
    Real,
    Boolean,
    CharacterString,
    Enumerated,
    Unsigned,
    Atomic,
)
from bacpypes.object import (
    Object,
    DeviceObject,
    AnalogValueObject,
    AnalogInputObject,
    AnalogOutputObject,
    BinaryValueObject,
    Property,
    register_object_type,
)

#
#   Proprietary Objects and their attributes
#

JCIDeviceObject = {
    "name": "JCIDeviceObject",
    "vendor_id": 5,
    "objectType": "device",
    "bacpypes_type": DeviceObject,
    "properties": {
        "SupervisorOnline": {"obj_id": 3653, "primitive": Boolean, "mutable": True},
        "Model": {"obj_id": 1320, "primitive": CharacterString, "mutable": True},
        "ModelType": {"obj_id": 32527, "primitive": CharacterString, "mutable": True},
        "State": {"obj_id": 2390, "primitive": CharacterString, "mutable": False},
        "MemoryUsage": {"obj_id": 2581, "primitive": Real, "mutable": False},
        "ObjectMemoryUsage": {"obj_id": 2582, "primitive": Real, "mutable": False},
        "CPU": {"obj_id": 2583, "primitive": Real, "mutable": False},
        "FlashUsage": {"obj_id": 2584, "primitive": Real, "mutable": False},
        "JCISystemStatus": {"obj_id": 847, "primitive": Enumerated, "mutable": False},
        "SABusPerformance": {
            "obj_id": 12157,
            "primitive": Enumerated,
            "mutable": False,
        },
        "SABusTokenLoopTime": {
            "obj_id": 12158,
            "primitive": Unsigned,
            "mutable": False,
        },
        "SABusCOVRcvPerMinute": {
            "obj_id": 12159,
            "primitive": Unsigned,
            "mutable": False,
        },
        "SABusCOVWritesPerMinute": {
            "obj_id": 12160,
            "primitive": Unsigned,
            "mutable": False,
        },
        "CPU Idle": {"obj_id": 30082, "primitive": Real, "mutable": False},
        "alarm": {"obj_id": 673, "primitive": Boolean, "mutable": False},
        "end_of_line": {"obj_id": 603, "primitive": Boolean, "mutable": False},
        "objectStatus": {"obj_id": 512, "primitive": Enumerated, "mutable": False},
    },
}
# EOL ?
# MEMORY ?

JCIAnalogValueObject = {
    "name": "JCIAnalogValueObject",
    "vendor_id": 5,
    "objectType": "analogValue",
    "bacpypes_type": AnalogValueObject,
    "properties": {
        "FLOW-SP_EEPROM": {"obj_id": 3113, "primitive": Real, "mutable": True},
        "Offset": {"obj_id": 956, "primitive": Real, "mutable": True},
        "Offline": {"obj_id": 913, "primitive": Boolean, "mutable": False},
        "SABusAddr": {"obj_id": 3645, "primitive": Unsigned, "mutable": False},
        "PeerToPeer": {"obj_id": 748, "primitive": Atomic, "mutable": False},
        "P2P_ErrorStatus": {"obj_id": 746, "primitive": Enumerated, "mutable": False},
    },
}

JCIAnalogInputObject = {
    "name": "JCIAnalogInputObject",
    "vendor_id": 5,
    "objectType": "analogInput",
    "bacpypes_type": AnalogInputObject,
    "properties": {
        "Offset": {"obj_id": 956, "primitive": Real, "mutable": True},
        "Offline": {"obj_id": 913, "primitive": Boolean, "mutable": False},
        "SABusAddr": {"obj_id": 3645, "primitive": Unsigned, "mutable": False},
        "InputRangeLow": {"obj_id": 1293, "primitive": Real, "mutable": True},
        "InputRangeHigh": {"obj_id": 1294, "primitive": Real, "mutable": True},
        "OutputRangeLow": {"obj_id": 1295, "primitive": Real, "mutable": True},
        "OutputRangeHigh": {"obj_id": 1296, "primitive": Real, "mutable": True},
    },
}

JCIAnalogOutputObject = {
    "name": "JCIAnalogOutputObject",
    "vendor_id": 5,
    "objectType": "analogOutput",
    "bacpypes_type": AnalogOutputObject,
    "properties": {
        "Offline": {"obj_id": 913, "primitive": Boolean, "mutable": False},
        "SABusAddr": {"obj_id": 3645, "primitive": Unsigned, "mutable": False},
        "InputRangeLow": {"obj_id": 1293, "primitive": Real, "mutable": True},
        "InputRangeHigh": {"obj_id": 1294, "primitive": Real, "mutable": True},
        "OutputRangeLow": {"obj_id": 1295, "primitive": Real, "mutable": True},
        "OutputRangeHigh": {"obj_id": 1296, "primitive": Real, "mutable": True},
        "polarity": {"obj_id": "polarity", "primitive": Enumerated, "mutable": True},
        "stroketime": {"obj_id": 3478, "primitive": Real, "mutable": True},
    },
}


def tec_short_point_list():
    return [
        ("binaryInput", 30827),
        ("binaryInput", 30828),
        ("binaryOutput", 86908),
        ("binaryOutput", 86909),
        ("binaryOutput", 86910),
        ("binaryOutput", 86911),
        ("binaryOutput", 86912),
        ("binaryOutput", 87101),
        ("binaryOutput", 87102),
        ("multiStateValue", 29501),
        ("multiStateValue", 29500),
        ("multiStateValue", 29509),
        ("multiStateValue", 29517),
        ("multiStateValue", 29518),
        ("multiStateValue", 29519),
        ("multiStateValue", 29520),
        ("multiStateValue", 29524),
        ("multiStateValue", 29525),
        ("multiStateValue", 29527),
        ("multiStateValue", 29712),
        ("multiStateValue", 29700),
        ("multiStateValue", 29709),
        ("multiStateValue", 29708),
        ("analogValue", 29505),
        ("analogValue", 29502),
        ("analogValue", 29503),
        ("analogValue", 29504),
        ("analogValue", 29506),
        ("analogValue", 29507),
        ("analogValue", 29508),
        ("analogValue", 29515),
        ("analogValue", 29522),
        ("analogValue", 29529),
        ("analogValue", 29530),
        ("analogValue", 29532),
        ("analogValue", 29701),
        ("analogValue", 29703),
        ("analogValue", 29705),
        ("analogValue", 29706),
        ("analogValue", 29707),
        ("analogValue", 29714),
        ("analogValue", 29717),
        ("analogValue", 29725),
        ("analogValue", 29726),
        ("analogValue", 29727),
        ("analogOutput", 86905),
        ("analogOutput", 86914),
        ("analogOutput", 86915),
        ("multiStateValue", 6),
        ("trendLog", 101010),
    ]



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
from . import Base
from . import Lite

try:
    from . import Complete
except ImportError:
    _PANDAS = False



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Simulate.py - simulate the value of controller I/O values
"""

# --- standard Python modules ---
# --- 3rd party modules ---
# --- this application's modules ---
from .IOExceptions import (
    OutOfServiceNotSet,
    OutOfServiceSet,
    NoResponseFromController,
    ApplicationNotStarted,
)


# ------------------------------------------------------------------------------


class Simulation:
    """
    Global informations regarding simulation
    """

    def sim(self, args):
        """
        Simulate I/O points by setting the Out_Of_Service property, then doing a 
        WriteProperty to the point's Present_Value.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        # with self.this_application._lock: if use lock...won't be able to call read...
        args = args.split()
        addr, obj_type, obj_inst, prop_id, value = args[:5]

        if self.read("{} {} {} outOfService".format(addr, obj_type, obj_inst)):
            self.write(
                "{} {} {} {} {}".format(addr, obj_type, obj_inst, prop_id, value)
            )
        else:
            try:
                self.write(
                    "{} {} {} outOfService True".format(addr, obj_type, obj_inst)
                )
            except NoResponseFromController as e:
                self._log.warning(
                    "Failed to write to OutOfService property ({})".format(e)
                )

            try:
                if self.read("{} {} {} outOfService".format(addr, obj_type, obj_inst)):
                    self.write(
                        "{} {} {} {} {}".format(
                            addr, obj_type, obj_inst, prop_id, value
                        )
                    )
                else:
                    raise OutOfServiceNotSet()
            except NoResponseFromController as e:
                self._log.warning(
                    "Failed to write to OutOfService property ({})".format(e)
                )

    def out_of_service(self, args):
        """
        Set the Out_Of_Service property so the Present_Value of an I/O may be written.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        # with self.this_application._lock: if use lock...won't be able to call read...
        args = args.split()
        addr, obj_type, obj_inst = args[:3]
        try:
            self.write("{} {} {} outOfService True".format(addr, obj_type, obj_inst))
        except NoResponseFromController as e:
            self._log.warning("Failed to write to OutOfService property ({})".format(e))

    def release(self, args):
        """
        Set the Out_Of_Service property to False - to release the I/O point back to 
        the controller's control.

        :param args: String with <addr> <type> <inst>

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        args = args.split()
        addr, obj_type, obj_inst = args[:3]
        try:
            self.write("{} {} {} outOfService False".format(addr, obj_type, obj_inst))
        except NoResponseFromController as e:
            self._log.warning("Failed to write to OutOfService property ({})".format(e))

        try:
            if self.read("{} {} {} outOfService".format(addr, obj_type, obj_inst)):
                raise OutOfServiceSet()
            else:
                pass  # Everything is ok"
        except NoResponseFromController as e:
            self._log.warning("Failed to read OutOfService property ({})".format(e))
